use embedded_hal::{
    digital::v2::{InputPin, OutputPin},
    timer::CountDown,
};

use fixed::types::extra::U2;

use fugit::{ExtU64, MicrosDurationU64};

use nb::block;

pub type FixedU16 = fixed::FixedU16<U2>;

#[derive(Debug)]
pub enum Error {
    IOError,
    OpenThermocouple,
}

pub struct Max6675<CS, SCK, SO>
where
    CS: OutputPin,
    SCK: OutputPin,
    SO: InputPin,
{
    cs: CS,
    sck: SCK,
    so: SO,
}

impl<CS, SCK, SO> Max6675<CS, SCK, SO>
where
    CS: OutputPin,
    SCK: OutputPin,
    SO: InputPin,
{
    pub fn new(cs: CS, sck: SCK, so: SO) -> Self {
        Self { cs, sck, so }
    }

    pub fn read_temperature<CD>(&mut self, timer: &mut CD) -> Result<FixedU16, Error>
    where
        CD: CountDown,
        CD::Time: From<MicrosDurationU64>,
    {
        self.cs.set_low().map_err(|_| Error::IOError)?;
        delay(timer, 10.micros());

        let mut d = 0u16;

        for i in (0..16).rev() {
            self.sck.set_low().map_err(|_| Error::IOError)?;
            delay(timer, 10.micros());

            if self.so.is_high().map_err(|_| Error::IOError)? {
                d |= 1 << i;
            }

            self.sck.set_high().map_err(|_| Error::IOError)?;
            delay(timer, 10.micros());
        }

        self.cs.set_high().map_err(|_| Error::IOError)?;

        if d & 0x4 == 0x4 {
            Err(Error::OpenThermocouple)
        } else {
            Ok(FixedU16::from_bits(d >> 3))
        }
    }
}

fn delay<CD>(timer: &mut CD, t: MicrosDurationU64)
where
    CD: CountDown,
    CD::Time: From<MicrosDurationU64>,
{
    timer.start(t);
    block!(timer.wait()).unwrap();
}
